//
// Virtual Trackball based on Quaternions
//
#include "Trackball.h"

#include <math.h>

//initializes trackball module
Trackball::Trackball(GLuint button) {
	GLint current_mode;

	tb_button = button;

	tb_button  = GLUT_MIDDLE_BUTTON;  //default mouse button for rotation
	tb_animate = GL_FALSE;  //flag indicating whether animation is enabled
	tb_rotate  = GL_FALSE;  //flag denoting ongoing rotation

	tb_lastquat[0] = 0.0;
	tb_lastquat[1] = 0.0;
	tb_lastquat[2] = 0.0;
	tb_lastquat[3] = 1.0;    

	radius         = 4.0;
	execute_idle   = false;

	//an easy way to load identity matrix
	//to tb_matrix ... ;-)
	glGetIntegerv(GL_MATRIX_MODE, &current_mode);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glGetFloatv(GL_MODELVIEW_MATRIX, &tb_matrix[0][0]);
	glPopMatrix();
	glMatrixMode(current_mode);
}

Trackball::~Trackball(void)
{
}

//internal vector functions
void Trackball::_VEC_copy(const GLfloat *v1, GLfloat *v2)
//copies vector data from v1 to v2
{
	register int i;

	for (i=0;i<3;i++) v2[i] = v1[i];
}

void Trackball::_VEC_scale(GLfloat *v, GLfloat s)
//scales vector v with scalar value s
{
	v[0] *= s;
	v[1] *= s;
	v[2] *= s;
}

void Trackball::_VEC_add(const GLfloat *v1, const GLfloat *v2, GLfloat *s)
//adds v1 to v2 and returns result in s
{
	s[0] = v1[0] + v2[0];
	s[1] = v1[1] + v2[1];
	s[2] = v1[2] + v2[2];
}

GLfloat Trackball::_VEC_dot(const GLfloat *v1, const GLfloat *v2)
//returns dot product of vectors v1 and v2
{
	return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
}

void Trackball::_VEC_cross(const GLfloat *v1, const GLfloat *v2, GLfloat *cross)
//calculates cross product of vectors v1 and v2 and returns
//result in parameter cross
{
	GLfloat t[3];

	t[0] = (v1[1]*v2[2]) - (v1[2]*v2[1]);
	t[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);
	t[2] = (v1[0]*v2[1]) - (v1[1]*v2[0]);
	_VEC_copy(t, cross);
}

void Trackball::_VEC_project(GLfloat v[3], int x, int y)
//projects the mouse position x,y to virtual trackball
//and returns the 3D position in vector v
{
	GLfloat x_coord;
	GLfloat y_coord;

	//normalize both x,y to be in [-1.0,1.0]
	//!! the upper left corner of a window handled
	//   by GLUT is (0,0) - this is not consistent
	//   with OpenGL... ;-(
	//   (read section "1.5 Conventions" on GLUT API web page)
	x_coord = (2.0 * x) / (GLfloat) tb_width - 1.0;
	y_coord = 1.0 - (2.0 * y) / (GLfloat) tb_height;

	//make sure x_coord and y_coord are within
	//[-1.0,1.0] after the above calculations
	if (x_coord > 1.0)
		x_coord = 1.0;
	else if (x_coord < -1.0)
		x_coord = -1.0;

	if (y_coord > 1.0)
		y_coord = 1.0;
	else if (y_coord < -1.0)
		y_coord = -1.0;

	//calculate z coordinate using TRACKBALL_RADIUS defined
	//in trackball.h and set components of v to be the calculated
	//values
	v[0] = x_coord;
	v[1] = y_coord;
	v[2] = sqrt (radius * radius - x_coord * x_coord - y_coord * y_coord);
}


//internal quaternion functions
void Trackball::_QUAT_norm(GLfloat quat[4])
//normalizes the passed quaternion quat
{
	register int i;
	GLfloat magnitude;

	//a unit quaternion is normalized just
	//like a unit vector...
	magnitude = sqrt (quat[0]*quat[0] + quat[1]*quat[1] + quat[2]*quat[2] + quat[3]*quat[3]);
	for (i = 0; i < 4; i++) {
		quat[i] /= magnitude;
	}
}

void Trackball::_QUAT_mult(GLfloat quat1[4], GLfloat quat2[4], GLfloat dest[4])
//multiplicates quaternions quat1 and quat2 and copies
//result to dest
//!! make sure a caller can use the same parameters
//   for quat2 and dest ...
//   (e.g. see invocation of this function in motion)
{
	GLfloat t1[4], t2[4], t3[4], tf[4];

	//multiply the things using the given formula
	//and using the given vector functions in this module

   _VEC_copy (quat1, t1);
 
   _VEC_scale (t1, quat2[3]);
 
   _VEC_copy (quat2, t2);
   _VEC_scale (t2, quat1[3]);
 
   _VEC_cross (quat1, quat2, t3);
 
   _VEC_add (t1, t2, tf);
   _VEC_add (t3, tf, tf);
 
   tf[3] = quat1[3] * quat2[3] - _VEC_dot (quat1, quat2);

	//copy result to dest
   dest[0] = tf[0];
   dest[1] = tf[1];
   dest[2] = tf[2];
   dest[3] = tf[3];

	//normalize calculated quaternion
	_QUAT_norm(dest);
}

void Trackball::_QUAT_buildQuat(GLfloat quat[4], GLfloat v[3])
//build new quaternion with given new position (parameter v)
//and the last position stored in 'tb_lastpos'
//return calculated quaternion in parameter quat
{
	GLfloat a[3];
	GLfloat dist[3];
	GLfloat phi;

	//return [0,0,0,1]-quaternion if mouse did not move
	//(i.e. new and last positions are identical)
	if (v[0] == tb_lastpos[0] && v[1] == tb_lastpos[1]) {
		quat[0] = quat[1] = quat[2] = 0;
		quat[3] = 1;
		return;
	}

	//calculate distance between new and last
	//3D position on virtual trackball
	dist[0] = v[0] - tb_lastpos[0];
	dist[1] = v[1] - tb_lastpos[1];
	dist[2] = v[2] - tb_lastpos[2];

	//calculate angle phi with the above calculated
	//distance (using again TRACKBALL_RADIUS)
	//!! make sure the value passed to function 'asin'
	//   lives within [-1.0,1.0] ...
	phi =
		sqrt (dist[0] * dist[0] + dist[1] * dist[1] +
		dist[2] * dist[2]) / (2.0 * radius);
	if (phi > 1.0)
		phi = 1.0;
	if (phi < -1.0)
		phi = -1.0;
	phi = 2.0 * asin (phi);

	//calculate rotation axis
	_VEC_cross (tb_lastpos, v, a);

	//normalize rotation axis
	_VEC_scale (a, 1.0 / (a[0] * a[0] + a[1] * a[1] + a[2] * a[2]));

	//build quaternion in parameter quat
	_VEC_copy (a, quat);
	_VEC_scale (quat, sin (phi / 2.0));
	quat[3] = cos (phi / 2.0);
}


void Trackball::_QUAT_buildMatrix(GLfloat quat[4])
//builds rotation matrix with given quaternion
{
	tb_matrix[0][0] = 1.0 - 2.0 * (quat[1]*quat[1] + quat[2]*quat[2]);
	tb_matrix[1][0] =       2.0 * (quat[0]*quat[1] - quat[2]*quat[3]);
	tb_matrix[2][0] =       2.0 * (quat[2]*quat[0] + quat[1]*quat[3]);
	tb_matrix[3][0] = 0.0;

	tb_matrix[0][1] =       2.0 * (quat[0]*quat[1] + quat[2]*quat[3]);
	tb_matrix[1][1] = 1.0 - 2.0 * (quat[2]*quat[2] + quat[0]*quat[0]);
	tb_matrix[2][1] =       2.0 * (quat[1]*quat[2] - quat[0]*quat[3]);
	tb_matrix[3][1] = 0.0;

	tb_matrix[0][2] =       2.0 * (quat[2]*quat[0] - quat[1]*quat[3]);
	tb_matrix[1][2] =       2.0 * (quat[1]*quat[2] + quat[0]*quat[3]);
	tb_matrix[2][2] = 1.0 - 2.0 * (quat[1]*quat[1] + quat[0]*quat[0]);
	tb_matrix[3][2] = 0.0;

	tb_matrix[0][3] = 0.0;
	tb_matrix[1][3] = 0.0;
	tb_matrix[2][3] = 0.0;
	tb_matrix[3][3] = 1.0;
}


// idle callback to be registered when doing animation
void Trackball::idle_callback(void) {
	if (execute_idle) {
		_QUAT_mult(tb_newquat, tb_lastquat, tb_lastquat);
		glutPostRedisplay();
	}
}

void Trackball::_start(int x, int y, GLint time) {
// initialize and start rotation
//  (see mouse)
	tb_timestamp = time;
	tb_rotate = GL_TRUE;

	_VEC_project(tb_lastpos, x, y);
}

void Trackball::_stop(GLint time)
// stoppes rotation and registeres or de-registeres
// idle callback
//  (see mouse)
{
	tb_rotate = GL_FALSE;
	if (time==tb_timestamp && tb_animate) execute_idle = true;
	else if (tb_animate) execute_idle = false;
}

void Trackball::animate(GLboolean animate) {
// sets animation flag
	tb_animate = animate;
}

void Trackball::mouse(int button, int state, int x, int y)
//reacts upon mouse stats
//(to be called from mouse callback function)
{
	if (button==tb_button && state==GLUT_DOWN) _start(x, y, glutGet(GLUT_ELAPSED_TIME));
	else if (button==tb_button && state==GLUT_UP) _stop(glutGet(GLUT_ELAPSED_TIME));
}

void Trackball::motion(int x, int y)
//reacts upon motion stats
//(to be called from motion callback function)
{
	GLfloat new_pos[3];

	//return immediately if rotation is not enabled
	if (!tb_rotate) return;

	//project current mouse position to virtual trackball
	_VEC_project(new_pos, x, y);
	//build new quaternion
	_QUAT_buildQuat(tb_newquat, new_pos);
	//save new position in 'tb_last'
	_VEC_copy(new_pos, tb_lastpos);
	//multiply new quaternion to last one
	_QUAT_mult(tb_newquat, tb_lastquat, tb_lastquat);

	//get time stamp
	tb_timestamp = glutGet(GLUT_ELAPSED_TIME);
	glutPostRedisplay();
}

void Trackball::reshape(int w, int h)
//stores new window size in internal variables
//(to be called from reshape callback function)
{
	tb_width = w;
	tb_height = h;
}

void Trackball::matrix(void)
//writes current rotation matrix to modelview stack
//(to be called from display callback function)
{
	GLint current_mode;

	//build current rotation matrix
	_QUAT_buildMatrix(tb_lastquat);

	//multiply current rotation matrix to modelview stack
	glGetIntegerv(GL_MATRIX_MODE, &current_mode);
	glMatrixMode(GL_MODELVIEW);
	glMultMatrixf(&tb_matrix[0][0]);
	glMatrixMode(current_mode);
}
